from abc import ABC, abstractmethod
from pathlib import Path
from typing import List


class Context(ABC):
    """
    Abstract context class that provides the interface of a Context.

    A context defines all the different locations where content can be requested during the execution of MOLGENIS
    Commander.

    Also functions as a context manager so an instance of this class can be set as the current context using the 'with'
    syntax:

    >>> context = Context() # an implementation of a context
    >>> with context:
    >>>     # do something
    """

    @abstractmethod
    def get_scripts_folder(self) -> Path:
        pass

    @abstractmethod
    def get_backups_folder(self) -> Path:
        pass

    @abstractmethod
    def get_issues_folder(self) -> Path:
        pass

    @abstractmethod
    def get_history_file(self) -> Path:
        pass

    @abstractmethod
    def get_dataset_folders(self) -> List[Path]:
        pass

    @abstractmethod
    def get_resource_folders(self) -> List[Path]:
        pass

    @abstractmethod
    def get_git_folders(self) -> List[Path]:
        pass

    @abstractmethod
    def get_properties_file(self) -> Path:
        pass

    @abstractmethod
    def get_storage_file(self) -> Path:
        pass

    def __enter__(self):
        """Replace the previous context (if any) with this context."""
        self.previous_context = context_holder.get_context()
        context_holder.set_context(self)

    def __exit__(self, type_, value, traceback):
        """When exiting this context, replace the previous one (if any)."""
        context_holder.set_context(self.previous_context)


# delay import of _context_holder to prevent circular import errors
import mcmd.core.context._context_holder as context_holder



"""
Maps json items from the REST API v2 to value objects from the 'system' module.
"""

from mcmd.molgenis.system import Role, Group, User, RoleMembership


def map_to_role(role: dict) -> Role:
    group = None
    if 'group' in role:
        group = map_to_group(role['group'])

    return Role(id=role['id'],
                name=role['name'],
                label=role['label'],
                group=group)


def map_to_user(user: dict) -> User:
    return User(id=user['id'],
                username=user['username'])


def map_to_group(group: dict) -> Group:
    return Group(id=group['id'],
                 name=group['name'])


def map_to_role_membership(membership: dict) -> RoleMembership:
    role = membership['role']
    user = membership['user']
    return RoleMembership(id=membership['id'],
                          user=map_to_user(user),
                          role=map_to_role(role))



import questionary

from mcmd.config import config
from mcmd.core.errors import McmdError
from mcmd.io import io


def multi_choice(message, choices):
    raise_if_non_interactive(message)

    question = questionary.select(message, choices=choices)
    return _ask(question)


def checkbox(message, choices):
    raise_if_non_interactive(message)

    checks = [{'name': choice, 'value': idx} for idx, choice in enumerate(choices)]

    question = questionary.checkbox(message, choices=checks)

    answer_ids = _ask(question, _validate_empty_list)
    return [choices[idx] for idx in answer_ids]


def input_(message, required=False):
    raise_if_non_interactive(message)

    question = questionary.text(message)

    if required:
        return _ask(question, _validate_empty_input)
    else:
        return _ask(question)


def password(message, required=False):
    raise_if_non_interactive(message)

    question = questionary.password(message)

    if required:
        return _ask(question, _validate_empty_input)
    else:
        return _ask(question)


def confirm(message):
    raise_if_non_interactive(message)

    question = questionary.confirm(message, default=False)
    return _ask(question)


def _ask(question, validation=None):
    answer = question.ask()

    if answer is None:
        # None result means that PyInquirer caught an InterruptException
        exit(0)
    elif validation:
        result = validation(answer)
        if result is not True:
            io.error(result)
            return _ask(question, validation)

    return answer


def _validate_empty_list(answer):
    return 'You must choose at least one option.' if len(answer) == 0 else True


def _validate_empty_input(answer):
    return "This field can't be empty." if len(answer) == 0 else True


def raise_if_non_interactive(message: str):
    if config.is_loaded() and config.get('settings', 'non_interactive'):
        raise McmdError('User input required but running in non-interactive mode. Message: {}'.format(message),
                        info="Please specify your command more precisely or switch to interactive mode with 'mcmd "
                             "config set interactive'")



import re
from datetime import datetime, timedelta
from typing import Optional

import pkg_resources
import requests
from packaging.version import Version
from requests import RequestException

from mcmd.core import store
from mcmd.io import io


def check():
    """
    Checks if a new version is available. Checks once every 8 hours at most. If a new version is available, a message
    will be shown for every executed command until the update has been installed.
    """
    current = _current_version()

    if datetime.now() - timedelta(hours=8) > store.get_last_version_check():
        latest = _latest_version()

        if latest and latest > current:
            store.set_update_available(latest)

        store.set_last_version_check(datetime.now())

    latest = store.get_update_available()
    if latest:
        if latest <= current:
            # an upgrade has just been done
            store.set_update_available(None)
        else:
            _show_update_message(current, latest)


def _current_version() -> Version:
    return Version(pkg_resources.get_distribution('molgenis-commander').version)


def _latest_version() -> Optional[Version]:
    """
    Gets the latest version from PyPi. Because there is no API for this, we regex it from the 'simple' page.
    """
    try:
        response = requests.get('https://pypi.org/simple/molgenis-commander/')

        if response.status_code == 200:
            versions = re.findall(r'molgenis-commander-(\d+\.\d+\.\d+)', response.text)
            latest = max(versions, key=lambda v: Version(v))
            return Version(latest)
    except RequestException:
        return None


def _show_update_message(current: Version, latest: Version):
    io.warn(
        'A new version of MOLGENIS Commander is available: {}! You are using {}.'.format(latest, current))
    io.info("To upgrade, run 'pip install --upgrade molgenis-commander'")



from typing import FrozenSet

import attr
import jinja2
from jinja2 import Environment, meta

_env = Environment(autoescape=False)


@attr.s(frozen=True, auto_attribs=True)
class Template:
    """
    A container for a Jinja2 template. Exposes the variables found in the template through the variables property.
    """

    string: str
    _template: jinja2.Template = attr.ib(init=False, eq=False)
    variables: FrozenSet[str] = attr.ib(init=False)

    @_template.default
    def __set_template(self):  # NOSONAR method is not unused
        return _env.from_string(self.string)

    @variables.default
    def __set_variables(self) -> frozenset:  # NOSONAR method is not unused
        parsed_content = _env.parse(self.string)
        return frozenset(meta.find_undeclared_variables(parsed_content))

    def render(self, values: dict) -> str:
        return self._template.render(values)



from pathlib import Path
from typing import List

from mcmd.args.actions import ParseKeyValue
from mcmd.commands._registry import arguments
from mcmd.core.command import command, CommandType
from mcmd.core.context import context
from mcmd.core.errors import McmdError
from mcmd.io import io
from mcmd.script import script_runner
from mcmd.script.model.script import Script
from mcmd.script.options import ScriptOptions
from mcmd.script.parser import script_parser
from mcmd.script.parser.errors import InvalidScriptError, ScriptValidationError
from mcmd.utils import files


@arguments('run', CommandType.META)
def add_arguments(subparsers):
    p_run = subparsers.add_parser('run',
                                  help='run a commander script',
                                  description="The script syntax is described on the wiki page: "
                                              "https://github.com/molgenis/molgenis-tools-commander/wiki/Scripts")
    p_run.set_defaults(func=run,
                       write_to_history=False)
    p_run.add_argument('script',
                       type=str,
                       help='a script from the scripts folder or a path to a script (when using --from-path)')
    p_run.add_argument('--ignore-errors', '-i',
                       action='store_true',
                       help='let the script continue when one or more commands throw an error')
    p_run.add_argument('--hide-comments', '-c',
                       action='store_true',
                       help="don't print comments during script execution")
    p_run.add_argument('--from-line', '-l',
                       type=int,
                       default=1,
                       help="the line number to start the script at (will still run earlier lines that contain "
                            "value declarations needed for execution)")
    p_run.add_argument('--from-path', '-p',
                       action='store_true',
                       help="run a script from a path instead of the ~/.mcmd/scripts folder")
    p_run.add_argument('--with-arguments', '-a',
                       metavar="KEY=VALUE",
                       nargs='+',
                       default=dict(),
                       help="list of arguments to pass to the script as key/value pairs",
                       action=ParseKeyValue)
    p_run.add_argument('--dry', '-d',
                       action='store_true',
                       help='runs the script without actually executing any of the commands')


@command
def run(args):
    script_file = _get_script(args)
    lines = files.read_file_lines(script_file)

    script = _try_parse_script(lines)

    options = ScriptOptions(arguments=args.with_arguments,
                            dry_run=args.dry,
                            start_at=args.from_line,
                            log_comments=not args.hide_comments,
                            exit_on_error=not args.ignore_errors)
    script_runner.run(script, options)


def _try_parse_script(lines: List[str]) -> Script:
    try:
        return script_parser.parse(lines)
    except InvalidScriptError as e:
        _print_errors_and_exit(e.errors)


def _print_errors_and_exit(errors: List[ScriptValidationError]):
    io.error('The script contains errors')
    io.newline()

    errors.sort(key=lambda e: e.line_number)

    for error in errors:
        print(error.message)
        io.newline()

    exit(1)


def _get_script(args):
    if args.from_path:
        script = Path(args.script)
    else:
        script = context().get_scripts_folder().joinpath(args.script)
    if not script.exists():
        raise McmdError("The script {} doesn't exist".format(script))
    return script



from mcmd.commands._registry import arguments
from mcmd.core.command import command
from mcmd.io import ask
from mcmd.io import io
from mcmd.io.io import highlight
from mcmd.molgenis import api
from mcmd.molgenis.client import put
from mcmd.molgenis.resources import ensure_resource_exists, ResourceType
from mcmd.molgenis.security import security


# =========
# Arguments
# =========

@arguments('disable')
def add_arguments(subparsers):
    p_disable = subparsers.add_parser('disable',
                                      help='disable resources and functionality',
                                      description="run 'mcmd disable rls -h' to view the help for those sub-commands")
    p_disable_subparsers = p_disable.add_subparsers(dest='type', metavar='')

    p_disable_rls = p_disable_subparsers.add_parser('row-level-security',
                                                    aliases=['rls'],
                                                    help='disables row level security on an entity type')
    p_disable_rls.set_defaults(func=disable_rls,
                               write_to_history=True)
    p_disable_rls.add_argument('entity',
                               type=str,
                               help="the entity type to remove the row level security from")

    p_disable_language = p_disable_subparsers.add_parser('language',
                                                         help='disables a language')
    p_disable_language.set_defaults(func=disable_language,
                                    write_to_history=True)
    p_disable_language.add_argument('language',
                                    type=str,
                                    help="the language you want to disable, specified by the two letter code (e.g. "
                                         "'en')")


# =======
# Methods
# =======

@command
def disable_rls(args):
    if not ask.confirm('Are you sure you want to disable row level security on %s?' % args.entity):
        return

    io.start('Disabling row level security on entity type %s' % highlight(args.entity))

    ensure_resource_exists(args.entity, ResourceType.ENTITY_TYPE)
    security.disable_row_level_security(args.entity)


@command
def disable_language(args):
    io.start('Disabling language {}'.format(highlight(args.language)))
    url = api.rest1('sys_Language/{}/active'.format(args.language))
    put(url, 'false')



from typing import List, Dict, Set

import attr

from mcmd.script.model.lines import ParsedLine, Line
from mcmd.script.parser.errors import ScriptValidationError


@attr.s(auto_attribs=True)
class _ParseState:
    lines: List[ParsedLine] = attr.Factory(list)
    raw_lines: List[str] = attr.Factory(list)
    combined_lines: List[Line] = attr.Factory(list)
    errors: List[ScriptValidationError] = attr.Factory(list)
    dependencies: Dict[ParsedLine, Set[ParsedLine]] = attr.Factory(dict)



"""
Makes a user a member of a role or includes a group role in another role.

Differentiates between 'normal' roles and 'group' roles:
- Normal roles are roles that do not belong to a group. A user can be a member of more than one of these roles.
- Group roles are roles that belong to a group. A user can be a member of just one of a group's roles.

This command won't do anything if a user is already member of the specified role. In the case a user is already member
of another role within the same group, this command will ask for confirmation before updating it.
"""

import json
import textwrap
from argparse import RawDescriptionHelpFormatter
from typing import Optional, List
from urllib.parse import urljoin

from mcmd.commands._registry import arguments
from mcmd.core.command import command
from mcmd.core.compatibility import version
from mcmd.core.errors import McmdError
from mcmd.io import io, ask
from mcmd.io.io import highlight
from mcmd.molgenis import api
from mcmd.molgenis.client import post, get, put
from mcmd.molgenis.principals import to_role_name, get_principal_type_from_args, PrincipalType
from mcmd.molgenis.rest_api_v2_mapper import map_to_role, map_to_user, map_to_role_membership
from mcmd.molgenis.system import User, Group, Role, RoleMembership
from mcmd.molgenis.version import get_version
from mcmd.utils.time import timestamp


# =========
# Arguments
# =========

@arguments('make')
def add_arguments(subparsers):
    p_make = subparsers.add_parser('make',
                                   help='make a user or role member of a role',
                                   formatter_class=RawDescriptionHelpFormatter,
                                   description=textwrap.dedent(
                                       """
                                       Make a user or role member of a (group) role.  
                                       
                                       When no subject type is specified (with --user or --role), the Commander will 
                                       try to detect it. 
                                       
                                       Version notes: 
                                       - since MOLGENIS 8.1 it is possible to let roles include group roles (can be
                                         used to make role ANONYMOUS or USER member of a group).
                                       - since MOLGENIS 8.3 group and role names are case sensitive and need
                                         to be typed exactly as-is. (Before 8.3 all role names will be upper cased 
                                         automatically).
                                       
                                       example usage:
                                         mcmd make john Biobanks_MANAGER
                                         mcmd make jane GCC_EDITOR                                        
                                         mcmd make --user jane GCC_EDITOR
                                         mcmd make --role ANONYMOUS GCC_VIEWER
                                       """
                                   ))
    p_make.set_defaults(func=make,
                        write_to_history=True)
    p_make.add_argument('subject',
                        type=str,
                        help='the user or role')
    p_make.add_argument('target_role',
                        metavar='role',
                        type=str,
                        help='the role to make the subject a member of')
    p_make_subject = p_make.add_mutually_exclusive_group()
    p_make_subject.add_argument('--user', '-u',
                                action='store_true',
                                help='flag to specify that the subject is a user')
    p_make_subject.add_argument('--role', '-r',
                                action='store_true',
                                help='flag to specify that the subject is a role')


# =======
# Methods
# =======

@command
def make(args):
    role_name = to_role_name(args.target_role)
    role = _get_role(role_name)
    subject_type = _get_subject_type(args)
    if subject_type == PrincipalType.USER:
        user = _get_user(args.subject)
        if role.group:
            _make_member_of_group_role(user, role)
        else:
            _make_member_of_role(role, user)
    elif subject_type == PrincipalType.ROLE:
        subject = _get_role(args.subject)
        _include_group_role(subject, role)
    else:
        raise ValueError("Unknown principal type")


# noinspection PyUnusedLocal
@version('7.0.0')
def _get_subject_type(args) -> PrincipalType:
    """
    Prior to 8.1 roles can't be included with the identities API so the subject is always a user.
    """
    if args.role:
        raise McmdError(
            "Including group roles is only possible in MOLGENIS 8.3 and up (you are using {})".format(get_version()))
    else:
        return PrincipalType.USER


@version('8.1.0')
def _get_subject_type(args) -> PrincipalType:
    """
    From 8.1 and up, group roles can be included with the identities API so the subject can be a user or a role.
    """
    return get_principal_type_from_args(args, principal_name=args.subject)


def _make_member_of_group_role(user: User, role: Role):
    group = role.group
    membership = _get_group_membership(user, group)
    if membership:
        if membership.role.name == role.name:
            io.info('User {} is already {} of group {}'.format(highlight(user.username),
                                                               highlight(role.label),
                                                               highlight(group.name)))
        else:
            update = ask.confirm(
                'User {} is {} of group {}. Do you want to update his/her role to {}?'.format(user.username,
                                                                                              membership.role.label,
                                                                                              group.name,
                                                                                              role.label))
            if update:
                _update_group_role_membership(user, group, role)
    else:
        _add_group_role_membership(user, group, role)


def _make_member_of_role(role, user):
    if _is_member(user, role):
        io.info('User {} is already a member of role {}'.format(highlight(user.username), highlight(role.name)))
    else:
        _add_role_membership(user, role)


def _add_group_role_membership(user: User, group: Group, role: Role):
    io.start('Making user {} a member of role {}'.format(highlight(user.username), highlight(role.name)))
    url = api.member(group.name)
    post(url, data={'username': user.username, 'roleName': role.name})


def _update_group_role_membership(user: User, group: Group, role: Role):
    io.start('Making user {} a member of role {}'.format(highlight(user.username), highlight(role.name)))
    url = urljoin(api.member(group.name), user.username)
    put(url, data=json.dumps({'roleName': role.name}))


def _add_role_membership(user: User, role: Role):
    """
    Adds a membership manually because the identities API can't add memberships to non-group roles.
    """
    io.start('Making user {} a member of role {}'.format(highlight(user.username), highlight(role.name)))
    membership = {'user': user.id,
                  'role': role.id,
                  'from': timestamp()}
    data = {'entities': [membership]}
    post(api.rest2('sys_sec_RoleMembership'), data=data)


@version('8.1.0')
def _include_group_role(subject: Role, target_role: Role):
    if not target_role.group:
        raise McmdError('Role {} is not a group role'.format(target_role.name))
    if subject.name == target_role.name:
        raise McmdError("A role can't include itself")

    io.start('Including role {} in role {}'.format(highlight(target_role.name), highlight(subject.name)))
    include = {'role': target_role.name}
    put(api.role(target_role.group.name, subject.name), data=json.dumps(include))


def _get_group_membership(user: User, group: Group) -> Optional[RoleMembership]:
    group_roles = _get_group_roles(group)
    group_role_ids = [role.id for role in group_roles]

    memberships = get(api.rest2('sys_sec_RoleMembership'),
                      params={
                          'attrs': 'id,user(id,username),role(id,name,label,group(id,name))',
                          'q': "user=={};role=in=({});(to=='',to=ge={})".format(user.id, ','.join(group_role_ids),
                                                                                timestamp())
                      }).json()['items']

    if len(memberships) == 0:
        return None
    else:
        return map_to_role_membership(memberships[0])


def _get_group_roles(group: Group) -> List[Role]:
    roles = get(api.rest2('sys_sec_Role'),
                params={
                    'attrs': 'id,name,label,group(id,name)',
                    'q': 'group=={}'.format(group.id)
                }).json()['items']

    if len(roles) == 0:
        raise McmdError('No roles found for group {}'.format(group.name))

    return [map_to_role(role) for role in roles]


def _is_member(user: User, role: Role) -> bool:
    memberships = get(api.rest2('sys_sec_RoleMembership'),
                      params={
                          'attrs': 'id',
                          'q': "user=={};role=={};(to=='',to=ge={})".format(user.id, role.id, timestamp())
                      }).json()['items']

    return len(memberships) > 0


def _get_user(user_name: str) -> User:
    users = get(api.rest2('sys_sec_User'),
                params={
                    'attrs': 'id,username',
                    'q': 'username=={}'.format(user_name)
                }).json()['items']

    if len(users) == 0:
        raise McmdError('Unknown user {}'.format(user_name))
    else:
        return map_to_user(users[0])


def _get_role(role_name: str) -> Role:
    roles = get(api.rest2('sys_sec_Role'),
                params={
                    'attrs': 'id,name,label,group(id,name)',
                    'q': 'name=={}'.format(role_name)
                }).json()['items']

    if len(roles) == 0:
        raise McmdError('No role found with name {}'.format(role_name))
    else:
        return map_to_role(roles[0])



from typing import Optional

from mcmd.core.context.base_context import Context

_context: Optional[Context] = None


def set_context(new_context: Context):
    global _context
    _context = new_context


def get_context() -> Context:
    return _context



"""
Entry point for security related actions. Delegates to permission_manager.py or permissions_api.py depending on the
MOLGENIS version being used and the action being performed.
"""

from mcmd.core.compatibility import version
from mcmd.core.errors import McmdError
from mcmd.molgenis.principals import PrincipalType
from mcmd.molgenis.resources import ResourceType
from mcmd.molgenis.security import permission_manager, permissions_api
from mcmd.molgenis.security.permission import Permission
from mcmd.molgenis.version import get_version


@version('7.0.0')
def grant_permission(principal_type: PrincipalType,
                     principal_name: str,
                     resource_type: ResourceType,
                     entity_type_id: str,
                     permission: Permission):
    """Grants permission on one (non-row) resource to one principal."""

    permission_manager.grant_permission(principal_type, principal_name, resource_type, entity_type_id, permission)


# noinspection PyUnusedLocal
@version('7.0.0')
def grant_row_permission(principal_type: PrincipalType,
                         principal_name: str,
                         entity_type_id: str,
                         entity_id: str,
                         permission: Permission):
    raise McmdError(
        "Giving row level security is only possible since MOLGENIS 8.1.1 (you are using {})".format(
            get_version()))


@version('8.1.1')
def grant_row_permission(principal_type: PrincipalType,
                         principal_name: str,
                         entity_type_id: str,
                         entity_id: str,
                         permission: Permission):
    """Grants permission on one row to one principal."""

    permissions_api.grant_row_permission(principal_type, principal_name, entity_type_id, entity_id, permission)


@version('7.0.0')
def enable_row_level_security(entity_type_id: str):
    """Enables row level security on an entity type."""

    permission_manager.enable_row_level_security(entity_type_id)


@version('7.0.0')
def disable_row_level_security(entity_type_id: str):
    """Disables row level security on an entity type."""

    permission_manager.disable_row_level_security(entity_type_id)



import attr


@attr.s(frozen=True, auto_attribs=True)
class ScriptOptions:
    arguments: dict = attr.Factory(dict)
    dry_run: bool = False
    start_at: int = 1
    log_comments: bool = True
    exit_on_error: bool = True



from abc import ABC, abstractmethod
from typing import List

from mcmd.script.model.lines import Line, ParsedLine


class ScriptValidationError(ABC, Exception):
    def __init__(self, line_number: int):
        self.line_number = line_number

    @property
    @abstractmethod
    def message(self):
        pass


class InvalidScriptError(Exception):
    def __init__(self, errors: List[ScriptValidationError]):
        self._errors = errors

    @property
    def errors(self) -> List[ScriptValidationError]:
        return self._errors


class ScriptSyntaxError(ScriptValidationError):
    """
    Raised when a line in the script contains a syntax error and couldn't be parsed.
    """

    def __init__(self, error_msg: str, line: Line, column: int = -1):
        super().__init__(line_number=line.number)
        self.error_msg = error_msg.capitalize()
        self.line = line
        self.column = column

    @property
    def message(self) -> str:
        msg = list()
        line_num_text = 'Line {}: '.format(self.line.number)
        msg.append(line_num_text + self.line.string)

        if self.column > -1:
            # add column indicator
            spacing = ' ' * len(line_num_text)
            spacing += ' ' * self.column
            msg.append(spacing + '^')

        msg.append("  - {}".format(self.error_msg))
        return '\n'.join(msg)


class ReassignmentError(ScriptValidationError):
    """
    Raised when a value has already been assigned in an earlier line.
    """

    def __init__(self, name: str, source: ParsedLine, other: ParsedLine):
        super().__init__(line_number=source.number)
        self._name = name
        self._source = source
        self._other = other

    @property
    def message(self):
        msg = list()
        msg.append('Line {}: {}'.format(self._source.number, self._source.raw))
        msg.append("  - Value '{}' is already assigned at line {}".format(self._name,
                                                                          self._other.number))
        return '\n'.join(msg)


class ForwardReferenceError(ScriptValidationError):
    """
    Raised when a Template references a value that is assigned at a later line.
    """

    def __init__(self, name: str, source: ParsedLine, referred: ParsedLine):
        super().__init__(line_number=source.number)
        self._name = name
        self._source = source
        self._referred = referred

    @property
    def message(self):
        msg = list()
        msg.append('Line {}: {}'.format(self._source.number, self._source.raw))
        msg.append("  - Value '{}' referenced before assignment: '{}' is assigned at "
                   "line {}".format(self._name,
                                    self._name,
                                    self._referred.number))
        return '\n'.join(msg)


class UnknownReferenceError(ScriptValidationError):
    """
    Raised when a Template references a value that is not assigned in the script.
    """

    def __init__(self, source: ParsedLine, reference: str):
        super().__init__(line_number=source.number)
        self._source = source
        self._reference = reference

    @property
    def message(self):
        msg = list()
        msg.append('Line {}: {}'.format(self._source.number, self._source.raw))
        msg.append("  - Unknown value '{}'".format(self._reference))
        return '\n'.join(msg)


class RecursiveReferenceError(ScriptValidationError):
    """
    Raised when a value is assigned that references itself.

    $value name = "{{name}}"
    """

    def __init__(self, source: ParsedLine, reference: str):
        super().__init__(line_number=source.number)
        self._source = source
        self._reference = reference

    @property
    def message(self):
        msg = list()
        msg.append('Line {}: {}'.format(self._source.number, self._source.raw))
        msg.append("  - Value '{}' referenced during assignment".format(self._reference))
        return '\n'.join(msg)



import argparse

import pkg_resources

from mcmd.args._formatter import GroupedHelpFormatter, list_subcommands_in_help
from mcmd.args._raising_parser import RaisingArgumentParser
# Import all commands:
# noinspection PyUnresolvedReferences
from mcmd.commands import *
from mcmd.commands import get_argument_adders
from mcmd.core.command import CommandType


# noinspection PyProtectedMember
def create_parser():
    parser = RaisingArgumentParser(prog='mcmd',
                                   usage=argparse.SUPPRESS,
                                   formatter_class=GroupedHelpFormatter,
                                   description=_description())
    subparsers = parser.add_subparsers(dest="command",
                                       metavar=""  # metavar is empty to hide the {add, delete, disable...} string
                                       )

    # add general optionals
    parser.add_argument('--as-user', '-u',
                        type=str,
                        metavar='USER',
                        help="execute a MOLGENIS command as a user")
    parser.add_argument('--with-password', '-p',
                        type=str,
                        metavar='PASSWORD',
                        help="the password to use when logging in")
    parser.add_argument('--verbose', '-v',
                        action='count',
                        help='print verbose messages')
    parser.add_argument('--version',
                        action='version',
                        version='MOLGENIS Commander {version}'.format(version=_get_version()))

    # Remove the 'commands' section title because we will be replacing it
    parser._positionals.title = None

    # add each command's arguments
    for type_ in CommandType:
        for argument_adder in get_argument_adders(type_):
            argument_adder(subparsers)

    # add each command's subcommands to its help string
    list_subcommands_in_help(parser)

    return parser


def _description():
    description = '''\
               _             _                     
     _____ ___| |___ ___ ___|_|___                 
    |     | . | | . | -_|   | |_ -|                
    |_|_|_|___|_|_  |___|_|_|_|___|      _         
           ___ _|___|___ _____ ___ ___ _| |___ ___ 
          |  _| . |     |     | .'|   | . | -_|  _|
          |___|___|_|_|_|_|_|_|__,|_|_|___|___|_|     v{}     


To read more about a (sub)command, run the command with the -h flag (for example: 'mcmd add -h', \
or 'mcmd add group -h').'''.format(_get_version())

    return description


def _get_version():
    return pkg_resources.get_distribution('molgenis-commander').version



from mcmd.script.model.lines import ParsedLine
from mcmd.script.model.statements import Templatable, Assignment
from mcmd.script.parser._parse_state import _ParseState
from mcmd.script.parser.errors import ReassignmentError, ForwardReferenceError, UnknownReferenceError, \
    RecursiveReferenceError


def resolve(state: _ParseState):
    """
    Resolves the dependencies of each ParsedLine and adds them to that ParsedLine. Attaches any errors encountered
    along the way to the parse state.
    """

    declarations_by_name = dict()
    for line in state.lines:
        statement = line.statement
        if isinstance(statement, Assignment):
            if statement.name in declarations_by_name:
                state.errors.append(ReassignmentError(statement.name, line, declarations_by_name[statement.name]))
            else:
                declarations_by_name[statement.name] = line

    for line in state.lines:
        statement = line.statement
        if isinstance(statement, Templatable):
            _set_dependencies(declarations_by_name, line, statement.variables, state)


def _set_dependencies(declarations_by_name: dict, line: ParsedLine, variables: frozenset, state: _ParseState):
    for var in variables:
        if var not in declarations_by_name:
            state.errors.append(UnknownReferenceError(line, var))
        else:
            dependency = declarations_by_name[var]

            if dependency.number == line.number:
                state.errors.append(RecursiveReferenceError(line, var))
            elif dependency.number >= line.number:
                state.errors.append(ForwardReferenceError(var, line, dependency))
            else:
                if line not in state.dependencies:
                    state.dependencies[line] = {declarations_by_name[var]}
                else:
                    state.dependencies[line].add(declarations_by_name[var])



"""
Contains bare-bones definitions of System Entity Types of MOLGENIS.
"""
from typing import Optional

import attr


@attr.s(frozen=True, auto_attribs=True)
class User:
    id: str
    username: str


@attr.s(frozen=True, auto_attribs=True)
class Group:
    id: str
    name: str


@attr.s(frozen=True, auto_attribs=True)
class Role:
    id: str
    name: str
    label: str
    group: Optional[Group] = None


@attr.s(frozen=True, auto_attribs=True)
class RoleMembership:
    id: str
    user: User
    role: Role



"""Client for the Permission API of MOLGENIS 8.1.1 and up."""
from urllib.parse import urljoin, quote

from mcmd.core.errors import McmdError
from mcmd.molgenis import api
from mcmd.molgenis.client import get, delete, post, patch
from mcmd.molgenis.principals import PrincipalType, to_role_name
from mcmd.molgenis.security.permission import Permission


def grant_row_permission(principal_type: PrincipalType,
                         principal_name: str,
                         entity_type_id: str,
                         entity_id: str,
                         permission: Permission):
    """Grants a permission on one row to one principal."""

    existing_permissions = _get_existing_permissions('entity-' + entity_type_id,
                                                     entity_id,
                                                     principal_type,
                                                     principal_name)

    if permission == Permission.NONE:
        if len(existing_permissions) != 0:
            _delete_permission(entity_type_id, entity_id, principal_type, principal_name)
    else:
        if len(existing_permissions) == 0:
            _add_permission(entity_type_id, entity_id, principal_type, principal_name, permission)
        else:
            _update_permission(entity_type_id, entity_id, principal_type, principal_name, permission)


def get_user_plugin_permission(plugin: str, user: str) -> Permission:
    """Gets the permission for a user on a plugin."""
    return _get_permission('plugin', plugin, PrincipalType.USER, user)


def get_role_plugin_permission(plugin: str, role: str) -> Permission:
    """Gets the permission for a role on a plugin."""
    return _get_permission('plugin', plugin, PrincipalType.ROLE, role)


def get_user_entity_type_permission(entity_type: str, user: str) -> Permission:
    """Gets the permission for a user on an entity type."""
    return _get_permission('entityType', entity_type, PrincipalType.USER, user)


def get_role_entity_type_permission(entity_type: str, role: str) -> Permission:
    """Gets the permission for a role on an entity type."""
    return _get_permission('entityType', entity_type, PrincipalType.ROLE, role)


def get_user_package_permission(package: str, user: str) -> Permission:
    """Gets the permission for a user on a package."""
    return _get_permission('package', package, PrincipalType.USER, user)


def get_role_package_permission(package: str, role: str) -> Permission:
    """Gets the permission for a role on a package."""
    return _get_permission('package', package, PrincipalType.ROLE, role)


def get_user_entity_permission(entity_type: str, entity, user: str) -> Permission:
    """Gets the permission for a user on an entity."""
    return _get_permission('entity-' + entity_type, entity, PrincipalType.USER, user)


def get_role_entity_permission(entity_type: str, entity, role: str) -> Permission:
    """Gets the permission for a role on an entity."""
    return _get_permission('entity-' + entity_type, entity, PrincipalType.ROLE, role)


def is_row_level_secured(entity_type_id: str) -> bool:
    """Returns whether the entity type is row level secured or not."""

    response = get(urljoin(api.permissions(), 'types'))
    types = response.json()['data']
    entity_type_ids = [t['entityType'] for t in types]
    return entity_type_id in entity_type_ids


def _get_permission(type_id: str, object_id, principal_type: PrincipalType, principal: str) -> Permission:
    object_id = str(object_id)
    url = urljoin(api.permissions(), '{}/{}?q={}=={}'.format(quote(type_id, safe=''),
                                                             quote(object_id, safe=''),
                                                             quote(principal_type.value),
                                                             quote(principal, safe='')))

    permissions = get(url).json()['data']['permissions']
    if len(permissions) == 0:
        return Permission.NONE
    else:
        return Permission[permissions[0]['permission']]


def _get_existing_permissions(type_id: str, object_id: str, principal_type: PrincipalType, principal_name: str):
    get_url = urljoin(api.permissions(),
                      '{}/{}?q={}=={}'.format(quote(type_id, safe=''),
                                              quote(object_id, safe=''),
                                              quote(principal_type.value),
                                              quote(principal_name, safe='')))

    existing_permissions = get(get_url).json()['data']['permissions']
    return existing_permissions


def _delete_permission(entity_type_id: str, entity_id: str, principal_type: PrincipalType, principal_name: str):
    url = _get_entity_url(entity_type_id, entity_id)
    body = _get_principal_body(principal_type, principal_name)
    delete(url, data=body)


def _add_permission(entity_type_id: str,
                    entity_id: str,
                    principal_type: PrincipalType,
                    principal_name: str,
                    permission: Permission):
    url = _get_entity_url(entity_type_id, entity_id)
    body = _get_principal_body(principal_type, principal_name)

    body['permission'] = permission.value.upper()
    post(url, data={'permissions': [body]})


def _update_permission(entity_type_id: str,
                       entity_id: str,
                       principal_type: PrincipalType,
                       principal_name: str,
                       permission: Permission):
    url = _get_entity_url(entity_type_id, entity_id)
    body = _get_principal_body(principal_type, principal_name)

    body['permission'] = permission.value.upper()
    patch(url, data={'permissions': [body]})


def _get_entity_url(entity_type_id: str, entity_id: str) -> str:
    return urljoin(api.permissions(),
                   'entity-{}/{}'.format(quote(entity_type_id, safe=''),
                                         quote(entity_id, safe='')))


def _get_principal_body(principal_type: PrincipalType, principal_name: str) -> dict:
    body = dict()
    if principal_type == PrincipalType.USER:
        body['user'] = principal_name
    elif principal_type == PrincipalType.ROLE:
        principal_name = to_role_name(principal_name)
        body['role'] = principal_name
    else:
        raise McmdError('Unknown principal type: %s' % principal_type)
    return body



from typing import List

from mcmd.script.model.lines import ParsedLine
from mcmd.script.model.script import Script
from mcmd.script.parser import dependency_resolver, line_parser, line_combiner
from mcmd.script.parser._parse_state import _ParseState
from mcmd.script.parser.errors import ScriptSyntaxError, InvalidScriptError


def parse(str_lines: List[str]) -> Script:
    """
    Parses a script and returns a Script object containing ParsedLines.
    """

    state = _ParseState()

    # first pass: strip lines
    state.raw_lines = [line.strip() for line in str_lines]

    # second pass: combine multi-line statements
    state.combined_lines = line_combiner.combine_multi_lines(state.raw_lines)

    # third pass: interpret the lines
    _parse_lines(state)

    # fourth pass: resolve dependencies
    dependency_resolver.resolve(state)

    if len(state.errors) > 0:
        raise InvalidScriptError(state.errors)
    else:
        return Script(lines=state.lines, dependencies=state.dependencies)


def _parse_lines(state: _ParseState):
    for line in state.combined_lines:
        try:
            statement = line_parser.parse(line)
        except ScriptSyntaxError as e:
            state.errors.append(e)
        else:
            state.lines.append(ParsedLine(raw=line.string, number=line.number, statement=statement))



from enum import Enum


class Permission(Enum):
    """Enum of permissions (and synonyms) that map to MOLGENIS permissions."""

    NONE = 'none'
    WRITEMETA = 'writemeta'
    READMETA = 'readmeta'
    EDIT = 'write'
    WRITE = 'write'
    VIEW = 'read'
    READ = 'read'
    COUNT = 'count'



from abc import abstractmethod
from typing import Optional, Any

import attr

from mcmd.script.model.templates import Template
from mcmd.script.model.types_ import InputType, ValueType


@attr.s(frozen=True)
class Statement:
    """A single or multi-lined statement in a MC script."""
    pass


@attr.s(frozen=True, auto_attribs=True)
class Assignment:
    """
    A statement that assigns a value. Should expose the value's name through the name property.

    $value type = "assignment"
    $input bool is_assigned
    """

    name: str


@attr.s(frozen=True)
class Templatable:
    """
    A statement that contains one or more Templates. Should expose all its Templates' variables through the variables
    property.
    """

    @property
    @abstractmethod
    def variables(self) -> frozenset:
        pass


@attr.s(frozen=True, auto_attribs=True)
class Value(Statement, Templatable, Assignment):
    """
    A value assignment. Supports text (in the form of a Template), booleans and None values.

    $value text = "text {{variable}}"
    $value bool = true
    $value empty
    """

    type: ValueType
    value: Any

    @classmethod
    def from_untyped_value(cls, name: str, value):
        if value is None:
            type_ = None
        elif isinstance(value, bool):
            type_ = ValueType.BOOL
        elif isinstance(value, Template):
            type_ = ValueType.TEMPLATE
        else:
            raise ValueError('invalid value type: {}'.format(type(value)))
        return cls(name, type_, value)

    @property
    def variables(self) -> set:
        if self.type == ValueType.TEMPLATE:
            return self.value.variables
        else:
            return set()


@attr.s(frozen=True, auto_attribs=True)
class Input(Statement, Templatable, Assignment):
    """
    An input assignment. Doesn't have a value because the inputs are requested from the user at runtime. Supports
    multiple input types and an optional message.

    $input bool yesno
    $input text type : "text"
    """

    type: InputType
    message: Optional[Template] = None

    @property
    def variables(self) -> frozenset:
        if self.message is not None:
            return self.message.variables
        else:
            return frozenset()


@attr.s(frozen=True, auto_attribs=True)
class Wait(Statement, Templatable):
    """
    A wait statement. Shows a message until the user presses enter.

    $wait : "Get a snack"
    """

    message: Template

    @property
    def variables(self) -> frozenset:
        return self.message.variables


@attr.s(frozen=True, auto_attribs=True)
class VisibleComment(Statement, Templatable):
    """
    Comments that will be printed to the output.

    # Creating user {{name}}
    """

    text: Template

    @property
    def variables(self) -> frozenset:
        return self.text.variables


@attr.s(frozen=True, auto_attribs=True)
class Command(Statement, Templatable):
    """
    A MOLGENIS Commander command.
    """

    command: Template

    @property
    def variables(self) -> frozenset:
        return self.command.variables


@attr.s(frozen=True, auto_attribs=True)
class InvisibleComment(Statement):
    """
    Comments that won't be processed in any way.

    // this is a comment
    """

    comment: str


@attr.s(frozen=True)
class Empty(Statement):
    """
    An empty line.
    """
    pass



class ArgumentSyntaxError(Exception):
    def __init__(self, message: str, usage: str):
        super().__init__(message)
        self._usage = usage

    @property
    def usage(self):
        return self._usage



import sys
from enum import Enum

from mcmd.args.errors import ArgumentSyntaxError
from mcmd.config import config
from mcmd.core import history
from mcmd.core.errors import McmdError, ScriptError
from mcmd.io import io
from mcmd.molgenis import auth


class CommandType(Enum):
    # Commands that communicate with a running MOLGENIS over HTTP
    STANDARD = 'standard'

    # Commands that communicate with the database, filestore, etc. of a locally installed MOLGENIS
    LOCAL = 'local'

    # Other commands like configuring MCMD, running scripts, seeing the history, etc.
    META = 'meta'


def command(func):
    """Decorator for commands. Handles auxiliary actions: authentication, history and errors."""

    def wrapper(args, nested=False):
        """
        :param args: the argparse arguments
        :param nested: if this command was started from another command
        :return: a decorated function
        """

        success = True
        try:
            _set_authentication(args)
            func(args)
        except ArgumentSyntaxError as error:
            # Some commands have additional argument checks that are evaluated at runtime
            success = False
            if nested:
                # let the encapsulating command deal with the error
                raise error
            else:
                _handle_syntax_error(error)
        except McmdError as error:
            success = False
            if nested:
                # let the encapsulating command deal with the error
                raise error
            else:
                _handle_error(error)
        else:
            io.succeed()
        finally:
            if args.write_to_history:
                history.write(args.arg_string, success=success)

    return wrapper


def _set_authentication(args):
    if args.as_user:
        if args.with_password:
            auth.set_(username=args.as_user, password=args.with_password, as_user=True)
        else:
            auth.set_(username=args.as_user, password=args.as_user, as_user=True)
    else:
        auth.set_(config.username(), config.password(), config.token())


def _handle_error(error):
    io.error(error.message)
    if error.info:
        io.info(error.info)

    if isinstance(error, ScriptError):
        io.info('Script failed at line {}'.format(io.highlight(str(error.line))))

    exit(1)


def _handle_syntax_error(error):
    sys.stderr.write(error.usage)
    sys.stderr.write(str(error))
    exit(1)



"""
All interactions with preset locations on the file system should go through a Context object. See the 'Context' class
in the 'base_context' module for more information, and see the 'HomeContext' class in the 'home_context' module for the
default implementation.

The current context should be retrieved by using the package method in this file.
"""

from mcmd.core.context import _context_holder
from mcmd.core.context.base_context import Context


def context() -> Context:
    return _context_holder.get_context()
